// jmode (XIM Server) currently for Japanese
// initialize many modules
//
#ifdef HAVE_CONFIG_H
# include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <ctype.h>
#include <stdlib.h>
#include <pwd.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include "jmode.h"

//
int g_option_mask;
int uid;
int scr_width,scr_height;
int host_byte_order;

//ʸ
#ifdef __FreeBSD__
#define LANG "ja_JP.EUC"
#else
#define LANG "ja_JP"
#endif
#define VERSION_NAME "jmode out of the way! Version "VERSION"\n"
char *version_name=VERSION_NAME;
char *usage=
"--help , --version :Show usage or version\n"
"--trace :trace-connection\n"
"--trace-xim :trace-xim-message\n";

static void
print_version()
{
    printf(version_name);
}

static void
print_usage()
{
    print_version();
    printf(usage);
    exit(0);
}

static void
parse_args(int argc, char **argv)
{
    int i;
    for (i = 1; i < argc; i++) {
	if (!strncmp(argv[i], "--", 2)) {
	    char *opt;
	    opt = &argv[i][2];
	    if (!strcmp(opt, "version")) {
		print_version();
		exit(0);
	    } else if (!strcmp(opt, "help")) {
		print_usage();
	    } else if(!strcmp(opt, "trace")) {
		g_option_mask |= OPT_TRACE;
	    } else if(!strcmp(opt, "trace-xim")) {
		g_option_mask |= OPT_TRACE_XIM;
	    }
	}
    }
}

static void
get_runtime_env()
{
    uid = getuid();

    int i = 1;
    char *v = (char *)&i;
    if (*v == 1) {
	host_byte_order = LSB_FIRST;
    } else {
	host_byte_order = MSB_FIRST;
    }
}

int
main(int argc, char **argv)
{
    int fd;

    putenv("LANG=" LANG);
    setlocale(LC_ALL, LANG);
    get_runtime_env();
    parse_args(argc, argv);
	
    if (!pretrans_register()) {
	return 1;
    }
    init_converter();
	
    connection_setup();
    fd = pretrans_setup();
    if (fd == -1) {
	return 0;
    }
    init_convdisp();

    main_loop();
    return 0;
}

int
pad4(int x)
{
    return (4-(x%4))%4;
}

void
hex_dump(unsigned char *buf, int len)
{
    int i, j;
    unsigned char c[16];
    for (i = 0 ; i < len ; i += 16) {
	printf("%8.8x ", i);
	for (j = 0 ; j < 16 ; j++) {
	    if (i + j < len) {
		c[j] = buf[i+j];
		if (j == 7) {
		    printf("%2.2x-", c[j]);
		} else {
		    printf("%2.2x ", c[j]);
		}
	    } else {
		c[j] = 0;
		printf("-- ");
	    }
	}
	printf(" ");
	for (j = 0; j < 16; j++) {
	    if (isprint(c[j])) {
		printf("%c", c[j]);
	    } else {
		printf(".");
	    }
	}
	printf("\n");
    }
}
/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 */
